/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */


#define  NEED_REPLIES       
#define    NEED_EVENTS

#include <X11/Xproto.h>
#include <X11/Xlib.h>
#include "vs.h"
#include "debug.h"
#include "xpacket.h"


static int debug=0;

VSHandleReplyOrEvent(protocol,real_root_window,new_root,request_status)
   BYTE        *protocol;
   Window      real_root_window;
   root_type   *new_root;
   int         request_status;
{
   if (*protocol == 1) {
      VSHandleReply(protocol,real_root_window,new_root,request_status);
   }
   else {
      VSHandleEvent(protocol,real_root_window,new_root);
   }
}

MatchAndSubstituteWindow(target_window,old_window,new_window)
   Window *target_window;
   Window old_window,
          new_window;
{

   DEBUG("++++++++++++++++++++++++++++++  MatchAndSubstituteWindow\n");
   if ( (*target_window) == old_window) {
      *target_window = new_window;
   }
}


VSHandleReply(reply,real_root_window,new_root,request_status)
   BYTE        *reply;
   Window      real_root_window;
   root_type   *new_root;
   int         request_status;
{

   DEBUG("VSHandleReply -- %s\n",request_name[request_status]);
   switch (request_status) {
      case X_GetGeometry :
           MatchAndSubstituteWindow( &( ((xGetGeometryReply*) reply)->root ) ,
                                     real_root_window, new_root->window);
           DEBUG(" real_root_window is 0x%08x, new_root_window is 0x%08x\n",
                 real_root_window,new_root->window);

           break;
      case X_QueryTree :
           MatchAndSubstituteWindow( &( ((xQueryTreeReply*) reply)->root ) ,
                                     real_root_window, new_root->window);
           break;
      case X_QueryPointer:
           MatchAndSubstituteWindow( &( ((xQueryPointerReply*) reply)->root ) ,
                                     real_root_window, new_root->window);
           VSAdjustRootXY( &(((xQueryPointerReply *) (reply))->rootX),
                           &(((xQueryPointerReply *) (reply))->rootY),
                           new_root->x, new_root->y);
           break;
   }
}

VSHandleEvent(event_packet,real_root_window,new_root)
   BYTE *event_packet;
   Window      real_root_window;
   root_type   *new_root;
{
   int     event_type;
   xEvent  *event;

   DEBUG("VSHandleEvent\n");
   event = (xEvent *) event_packet;
   switch (event-> u.u.type & 0177) {
      case KeyPress:
      case KeyRelease:
      case ButtonPress:
      case ButtonRelease:
      case MotionNotify:
      case EnterNotify:
          {
             MatchAndSubstituteWindow(&(event->u.keyButtonPointer.root),
                                      real_root_window, new_root->window);
             DEBUG("  Root window was corrected. \n");
             DEBUG("  Start root location correcting. \n");
             VSAdjustRootXY( &(event->u.keyButtonPointer.rootX),
                             &(event->u.keyButtonPointer.rootY),
                             new_root->x, new_root->y );
             DEBUG("  End   root location correcting. \n");
           }
           break;
      case LeaveNotify:
          {
             MatchAndSubstituteWindow(&(event->u.enterLeave.root),
                                      real_root_window, new_root->window);
             DEBUG("  Root window was corrected. \n");
             DEBUG("  Start root location correcting. \n");
             VSAdjustRootXY( &(event->u.enterLeave.rootX),
                             &(event->u.enterLeave.rootY) ,
                             new_root->x, new_root->y);
             DEBUG("  End   root location correcting. \n");
           }
           break;
    }
   event_type =  *((unsigned char *) event_packet);
   DEBUG(" Event -- %d\n", event_type);
   if (event_type >MAX_EVENT) {
      DEBUG(" Event -- Unknown event!!! -- event type # = %d\n",event_type);
   }
   else {
      DEBUG(" Event -- %s\n",event_name[event_type]);
   }
}


VSHandleRequestProtocol(request_protocol,filter)
   BYTE        *request_protocol;
   filter_type *filter;
{
   xReq       *x_request;
   static int request_count = 0;
   int        request_type;

   x_request = (xReq *) request_protocol;

   request_type = (int) (x_request->reqType);
   filter->request_status = request_type;
   if (request_type> 127) {
      DEBUG("Request length = %d, # = %d, name = Danger Extension\n",
             RequestLength(request_protocol),request_type);
   }
   else {
      DEBUG("Number %d request : Request length = %d, # = %d, name = %s\n",
            request_count++, RequestLength(request_protocol),
            request_type,request_name[request_type]);
   }
   if (RequestLength(request_protocol)>MAXIMUM_SIZE_OF_A_PROTOCOL) {
      printf("Request Protocol too big!!!\n");
      printf("  -- request protocol is %d bytes.\n",
             RequestLength(request_protocol));
      printf("  -- allowed size is %d bytes.\n",MAXIMUM_SIZE_OF_A_PROTOCOL);
      exit(0);
   }
}


VSHandleSetupRequest(setup_request)
   BYTE *setup_request;
{
}



VSHandleSetupReply(setup_reply,new_root)
   BYTE *setup_reply;
   root_type   *new_root;
{
   BYTE   *FindByteLocationOfRootWindowInSetupReply();
   Window *root_window;

   root_window=(Window *)FindByteLocationOfRootWindowInSetupReply(setup_reply);
   (*root_window) = new_root->window;
}


VSAdjustRootXY(x_to_root,y_to_root,root_at_x,root_at_y)
   INT16 *x_to_root,
         *y_to_root;
   int   root_at_x,
         root_at_y;
{

   DEBUG("Start CorrectRootXY()\n");


   /*
   DEBUG("-- Original rootX at %d, rootY at %d\n",*rootX,*rootY);
   DEBUG("-- New root window X0=  %d Y0= %d\n",*new_root_at_x,*new_root_at_y);
   DEBUG("-- New root window X0=  %d Y0= %d\n",
          (INT16)(*new_root_at_x),(INT16) (*new_root_at_y) );
   */

   *x_to_root =  *x_to_root - ((INT16) root_at_x);
   *y_to_root =  *y_to_root - ((INT16) root_at_y);

   /*
   DEBUG("-- New rootX= %d rootY= %d\n",*rootX,*rootY);
   */
}


