/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

/*
   htos.c
   Subscriber-end rotuines for the host-to-subscriber protocol

    Modifications :
                    FSH  24-SEP-91 Rewrote HTOS_AllocColor() and separated
                                   and added HTOS_AllocNamedColor()


*/

#include <stdio.h>

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "subscriber.h"
#include "tool.h"
#include "network.h"
#include "protocol.h"
#include "token.h"
#include "memory.h"
#include "misc.h"
#include "packet.h"
#include "htos.h"
#include "spacket.h"            /* FSH 4-SEP-91 */
#include "xpacket.h"            
/*
 * Begin Murray Addition
 */
#include "tokenInfo.h"
#include "particInfo.h"
/*
 * End Murray Addition
 */

#define   DEBUGHTOS_AddNewTool      0&DEBUGON
#define   DEBUGHTOS_DeleteTool      0&DEBUGON
#define   DEBUGHTOS_DroppedYou      0&DEBUGON
#define   DEBUGHTOS_X11Packet      0&DEBUGON
#define DEBUGHTOS_X11_PopUp_Packet   0&DEBUGON
#define   DEBUGHTOS_InternalAtom      0&DEBUGON
#define DEBUGHTOS_AllocColor            0&DEBUGON
#define DEBUGHTOS_AllocNamedColor       0&DEBUGON 
#define   DEBUGHTOS_TakeToken      0&DEBUGON
#define   DEBUGHTOS_ForceTokenRelease   0&DEBUGON
#define   DEBUGHTOS_AskTokenRelease   0&DEBUGON
#define DEBUGHTOS_YouAre      0&DEBUGON
#define   DEBUGHTOS_CreatorTakeToken   0&DEBUGON
#define DEBUGHTOS_ResizeWindow          0&DEBUGON
#define DEBUGHTOS_EndSession          0&DEBUGON

extern BYTE vs_in_out;
extern int has_reply[];
extern BYTE VERBOSE;
extern BYTE SEQNUM;
extern int OutputCount;
extern POPUP;
/* wang_x add*/
extern BYTE RESIZE;
/* end wang_x add*/

int CatchUp = 0;
BOOL ExpectPopUp = FALSE;
extern TimeOutFlag;
extern HostDisplayNumber;
extern BOOL SpecialQueryFlag;
extern BOOL HostQueryFlag;
#ifndef WINDOW
#define WINDOW  Window
#endif

#ifdef sun3
#define SUN
#endif

#ifdef sun4
#define SUN
#endif

#ifdef dec5000
#define DEC
#endif

#ifdef rs6000
#define IBM
#endif


/*
 *   HTOS_EndSession()
 *
 *   The chairman has ended the session.  Clean up NEATLY and
 *   exit.
 *
 *   Protocol:   HS_EndSession
 *
 *      HS_EndSession      Protocol ID Tag
 *
 *   RETURN VALUE:   void
 */
void HTOS_EndSession( Port, Who )
Socket Port;
Subscriber *Who;
{
   extern Widget   WSubscriber;
   extern Widget   WTool;
   Arg      args[5];
   int      n=0;
# if DEBUGHTOS_EndSession
   fprintf(stderr,"HTOS_EndSession()\n");
# endif

   XtSetArg( args[n],      XtNsensitive,           FALSE );
   n++;
   XtSetValues( WSubscriber, args, n );
   XtSetValues( WTool, args, n );

   CloseAllTools();

   UpdateMessage( "The Host has ended the session.\n Press [Exit XTV] to exit." );
}


/*
 *   HTOS_DroppedYou()
 *
 *   Somebody has booted you out.
 *
 *   Protocol:   HS_DroppedYou
 *
 *      HS_DroppedYou      Protocol ID Tag
 *
 *   RETURN VALUE:   void
 */
void HTOS_DroppedYou( Port, Who )
Socket Port;
Subscriber *Who;
{
   extern Widget   WSubscriber;
   extern Widget   WTool;
   Arg      args[5];
   int      n=0;

   BYTE            *TargetID;
   BYTE            *LeftOrEjected;
   Subscriber      *WhoItIs;

   TargetID = (BYTE *)ReadData( Port, sizeof(BYTE), BLOCK );

   if ( (WhoItIs=FindSubscriberByID(*TargetID)) != NULL ) {
#       if DEBUGHTOS_DroppedYou
      fprintf( stderr,"%s has been ejected  from the session\n", WhoItIs->LoginName);
#       endif
   } else {
#       if DEBUGHTOS_DroppedYou
      fprintf( stderr,"WARNING:  Subscriber ID %d unknown.", *TargetID);
#       endif
   }


   XtSetArg( args[n],      XtNsensitive,           FALSE );
   n++;
   XtSetValues( WSubscriber, args, n );
   XtSetValues( WTool, args, n );
   UpdateMessage( "You have been ejected from the session.\n Press [Exit XTV] to exit." );
   CloseAllTools();
}


/*
 *   HTOS_AddNewSubscriber()
 *
 *   Get a new subscriber structure from the host and add it to the list
 *
 *   Protocol:   HS_AddNewSubscriber | NewDude
 *
 *      HS_AddNewSubscriber      Protocol ID Tag
 *      NewDude         Subscriber Structure
 *
 *   RETURN VALUE:   void
 */
void HTOS_AddNewSubscriber( Port, Who )
Socket Port;
Subscriber *Who;
{
   Subscriber   *NewDude;

   NewDude = (Subscriber *)ReadData( Port, sizeof(Subscriber), BLOCK );

   AddSubscriber( NewDude );
   SetSubID( NewDude->SubscriberID, NewDude );

   UpdateMessage( "%s has joined the session.", NewDude->LoginName );
}


/*
 *   HTOS_DeleteSubscriber()
 *
 *   Cook a subscriber.
 *
 *   Protocol:   HS_DeleteSubscriber | TargetID | LeftOrEjected
 *
 *      HS_DeleteSubscriber   Protocol ID Tag
 *      TargetID      Subscriber ID of victim
 *      LeftOrEjected      00 - Left session
 *               01 - Ejected by chairman
 *
 *   RETURN VALUE:   void
 */
void HTOS_DeleteSubscriber( Port, Who )
Socket Port;
Subscriber *Who;
{
   BYTE      *TargetID;
   BYTE      *LeftOrEjected;
   Subscriber   *WhoItIs;

   TargetID = (BYTE *)ReadData( Port, sizeof(BYTE), BLOCK );
   LeftOrEjected = (BYTE *)ReadData( Port, sizeof(BYTE), BLOCK );

   if ( (WhoItIs=FindSubscriberByID(*TargetID)) != NULL ) {
      UpdateMessage( "%s has %s the session", WhoItIs->LoginName, ( *LeftOrEjected  ? "been ejected from" : "left" ) );
   } else {
      UpdateMessage( "WARNING:  Subscriber ID %d unknown.", *TargetID );
   }

   DeleteSubscriber( WhoItIs );
}



/*
 *   HTOS_AddNewTool()
 *
 *   Get a new tool structure from the host and add it to the list
 *
 *   Protocol:   HS_AddNewTool | VS_IN_OUT | NewTool | AuthData | VisualIDList
 *
 *      HS_AddNewTool      Protocol ID Tag
 *      NewTool         Tool Structure
 *      AuthData      Authorization data from connect request
 *      VisualIDList      List of Visual IDs
 *
 *   RETURN VALUE:   void
 */
void HTOS_AddNewTool( Port, Who )
Socket Port;
Subscriber *Who;
{
   Tool   *NewTool;
   BYTE   ByteOrder;

#if DEBUGHTOS_AddNewTool
   fprintf( stderr, "HTOS_AddNewTool(): Reading Tool Info\n" );
#endif


   NewTool = (Tool *)ReadData( Port, sizeof(Tool), BLOCK );

   vs_in_out = NewTool->vs_in_out;
   /*
   * Begin Murray Addition
   */

   /*
   * We don't want to mess with the subscriber list that the host maintains.
   */

   NewTool->subscrList = 0L;

   /*
   * End Murray Addition
   */

   NewTool->CatchUpMode = CatchUp;
   CatchUp = 0;

   ByteOrder = NewTool->ConnectPrefix.byteOrder;

   NewTool->HowMuchAuthData = (CARD16)SWAP_INT( NewTool->HowMuchAuthData, ByteOrder );
   NewTool->NumVisualIDs    = (CARD16)SWAP_INT( NewTool->NumVisualIDs,    ByteOrder );

#if DEBUGHTOS_AddNewTool
   fprintf( stderr, "HTOS_AddNewTool(): ToolCommand Name [%s]\n", (NewTool->ToolCommand));
   fprintf( stderr, "HTOS_AddNewTool(): ToolPID  [%d]\n", (NewTool->ToolPID));
   fprintf( stderr, "HTOS_AddNewTool(): ToolID  [%d]\n", (NewTool->ToolID));
   fprintf( stderr, "HTOS_AddNewTool(): ToolGID  [%d]\n", (NewTool->ToolGID));
   fprintf( stderr, "HTOS_AddNewTool(): Leader  [%d]\n", (NewTool->Leader));
   fprintf( stderr, "HTOS_AddNewTool(): Authorization Data is [%04x]\n", (NewTool->HowMuchAuthData) );
   fprintf( stderr, "HTOS_AddNewTool(): Authorization Data has [%d bytes]\n", (int)(NewTool->HowMuchAuthData) );
   fprintf( stderr, "HTOS_AddNewTool(): Tool Visual ID List is [%04x]\n", (NewTool->NumVisualIDs) );
   fprintf( stderr, "HTOS_AddNewTool(): Tool Visual ID List has [%d elements]\n", (int)(NewTool->NumVisualIDs) );
#endif

#if DEBUGHTOS_AddNewTool
   fprintf( stderr, "HTOS_AddNewTool(): Reading Tool Authorization Data [%d bytes]\n", (int)NewTool->HowMuchAuthData
       );
#endif

   NewTool->AuthData = ReadData( Port, NewTool->HowMuchAuthData, BLOCK );

#if DEBUGHTOS_AddNewTool
   fprintf( stderr, "HTOS_AddNewTool(): Reading Tool Visual ID List [%d elements]\n", (int)NewTool->NumVisualIDs );
#endif

   NewTool->VisualIDList = (XID *)ReadData( Port, ((NewTool->NumVisualIDs) * sizeof(XID)), BLOCK );

#if DEBUGHTOS_AddNewTool
   fprintf( stderr, "HTOS_AddNewTool(): The VisualIDList Contains %d items\n", NewTool->NumVisualIDs);
   fprintf( stderr, "HTOS_AddNewTool(): The VisualIDList elements are\n");

   {
      int i;
      i = NewTool->NumVisualIDs;
      while (i)
         printf ("%08x\n",NewTool->VisualIDList[NewTool->NumVisualIDs - i--]);
   }

#endif


   if ( ProcessClientConnection(NewTool) < 0 ) {
      /* xyzzy - Free contents of tool structure */
      return;
   }

#if DEBUGHTOS_AddNewTool
   fprintf( stderr, "HTOS_AddNewTool(): The Local VisualIDList Contains %d items\n", NewTool->LocalNumVisualIDs);
   fprintf( stderr, "HTOS_AddNewTool(): The Local VisualIDList elements are\n");
   {
      int i;
      i = NewTool->LocalNumVisualIDs;
      while (i)
         printf ("%08x\n",NewTool->LocalVisualIDList[NewTool->LocalNumVisualIDs - i--]);
   }

#endif

   (void) IntCircSeqMap(&(NewTool->seqMap));
   NewTool->ToolRequestNumber = 0;
   NewTool->ServerRequestNumber = 0;
   NewTool->LastSequenceNumber = 0;
   NewTool->InCount = 0;
   NewTool->OutCount = 0;
   SetToolID( NewTool );
   AddTool( NewTool );



   if (NewTool->Leader == TRUE)
      UpdateMessage( "Tool %s has just been connected.", NewTool->ToolCommand );

   /*
   * Begin Murray Addition
   */

   /*
   * Confirm the acceptance of the new tool. This tells the host to add me
   * to the list of people interested in this tool. Note that this should go
   * to a dialog box in the future. However, there will be a problem with the
   * code that follows, especially if more than one tool request is coming
   * down the pipe. We will want to pass the tool id as a parameter to the
   * dialog popup, so that we can later get the info.
   */

   if (SendByte(GetHostConnection(), SH_ConfirmNewTool) == ERROR)
      EndSession();
   if (SendByte(GetHostConnection(), NewTool->ToolID) == ERROR)
      EndSession();

   /*
   * End Murray Addition
   */

   if (NewTool->ToolCreator == GetLocalSubID()) {
      if (SendByte(GetHostConnection(), SH_GiveTokenToMe)==ERROR)
         EndSession();
      if (SendByte(GetHostConnection(), NewTool->ToolID)==ERROR)
         EndSession();
   }

   if( NewTool->CatchUpMode == 1 )
      if (SendByte(GetHostConnection(),SH_SendArchivePacket)==ERROR)
         EndSession();
}


/*
 *   HTOS_DeleteTool()
 *
 *   Cook a Tool.
 *
 *   Protocol:   HS_DeleteTool | TargetID
 *
 *      HS_DeleteTool      Protocol ID Tag
 *      TargetID      Tool ID of victim
 *      LeftOrEjected      00 - Left session
 *               01 - Ejected by chairman
 *
 *   RETURN VALUE:   void
 */
void HTOS_DeleteTool( Port, Who )
Socket Port;
Subscriber *Who;
{
   BYTE      ToolID;
   Tool      *WhichTool;

   ToolID = GetByte( Port);
#   if DEBUGHTOS_DeleteTool
   fprintf( stderr, "HTOS_DeleteTool(): ToolId to be deleted is: %d\n", 
       ToolID);
#   endif

   if ( (WhichTool=FindToolByID(ToolID)) == NULL ) {
#   if DEBUGHTOS_DeleteTool
      fprintf(stderr, "WARNING:  Tool ID %d unknown.", ToolID );
#   endif
      return;
   }
   if( WhichTool->CatchUpMode == 1 ) if (SendByte(GetHostConnection(),SH_SendArchivePacket)==ERROR) EndSession();
   CleanUpTool( WhichTool );

#       if DEBUGHTOS_DeleteTool
   fprintf( stderr, "HTOS_DeleteTool(): Cleaned Up \n");
#       endif


}


/*
 *   HTOS_X11Packet()
 *
 *   Cook a Tool.
 *
 *   Protocol:   HS_X11Packet | Tool ID | Length | Packet
 *
 *      HS_X11Packet      Protocol ID Tag
 *      ToolID         Tool ID of sender
 *      Length         32-bit packet length in byte-order of tool
 *      Packet         X11 Packet
 *
 *   RETURN VALUE:   void
 */
XWindowAttributes  Win_Attr;

void HTOS_X11Packet( Port, Who )
Socket Port;
Subscriber *Who;
{
   BYTE      *ToolID;
   Tool      *WhichTool;
   CARD32      *Length;
   BYTE      *Packet;
   xReq            *PHeader;
   INT16           temp;
   CARD32          BitMask;
   int             BitNumber;
   int             WhereAt = 0;
   int             abs_x,abs_y;
   Window          dummywin;
   Window          root;
   Window          parent;
   Window          children;
   unsigned        int     nchildren;
   int      returnVal;
/*
** Begin Kantsios Addition
*/
   CARD16          *TheSeqNum;
/*
** End Kantsios Addition
*/




   ToolID = (BYTE *)ReadData( Port, sizeof(BYTE), BLOCK );
   if ( (WhichTool=FindToolByID(*ToolID)) == NULL ) {
#       if DEBUGHTOS_X11Packet
      fprintf(stderr,  "X11Packet(): Tool ID %d unknown.", *ToolID );
#   endif
      FreeMem(ToolID);
      TrashIt(Port);
      return;
   }
   Length = (CARD32 *)ReadData( Port, sizeof(CARD32), BLOCK );
*Length = SWAP_LONG( *Length, WhichTool->ConnectPrefix.byteOrder );
   Packet = (BYTE *)ReadData( Port, *Length, BLOCK );
/*
** Begin Kantsios Addition
*/
if (WhichTool->seqnum){

   TheSeqNum = (CARD16 *)ReadData( Port, sizeof(CARD16), BLOCK );
   *TheSeqNum = ntohs( *TheSeqNum);
}

/*
** End Kantsios Addition
*/

   if ( WhichTool->verbose ){
      if ( ((++(WhichTool->InCount) + WhichTool->OutCount) % (10*OutputCount)) == 0){ 
		printf("\t%d -> %s -> %d\n",
			WhichTool->OutCount, WhichTool->ToolCommand, WhichTool->InCount);
		fflush(stdout);
      }
      else if ((WhichTool->InCount % OutputCount) == 0) write(1,".", 2);
   }


#       if DEBUGHTOS_X11Packet
   fprintf( stderr, "HTOS_X11Packet(): OpCode [%d], Length [%d], has_reply [%s]\n", *Packet, *Length+32,
       (has_reply[*Packet] ? "TRUE" : "FALSE"));
#       endif
   HandleSpecialQuery(WhichTool,*Packet);
   /*
        if (HostQueryFlag) {
#       if DEBUGHTOS_X11Packet
fprintf( stderr, "HTOS_X11Packet(): got SpecialHostQuery request: Trash it\n");
#       endif
fprintf( stderr, "HTOS_X11Packet(): got SpecialHostQuery request: Trash it\n");
        
        HostQueryFlag = FALSE;
        FreeMem( ToolID );
        FreeMem( Length );
        FreeMem( Packet );
        return;
        }
*/
   if ( TranslateRequest( Packet, WhichTool ) ) {
      /* debug
if (*Packet == 18){
fprintf( stderr, "HandleIncomingClientPacket(): Got Request [%d], pending_reply [%d]\n",
            *Packet,  has_reply[*Packet] );
fprintf( stderr," window atom atom %o8x %o8x %o8x\n", *((CARD32 *)Packet+4),
                *((CARD32 *)(Packet+8)),*((CARD32 *)(Packet+12)));
}
*/



      if ((WhichTool->popup)&&( *Packet == 120)) {
         ExpectPopUp = TRUE;

#       if DEBUGHTOS_X11_PopUp_Packet
         fprintf( stderr, "HTOS_X11Packet(): PopUpInfo Packet received \n");
#       endif
         XQueryTree(XtDisplay((Widget *)GetTopLevelWidget()),
             SWAP_LONG(*((Window *)(Packet+4)),WhichTool->ConnectPrefix.byteOrder),
             &root, &parent, &children, &nchildren);
         XGetWindowAttributes(XtDisplay((Widget *)GetTopLevelWidget()),
             SWAP_LONG(*((Window *)(Packet+4)),WhichTool->ConnectPrefix.byteOrder),
             &Win_Attr);
         if (root != parent) {
            XTranslateCoordinates(XtDisplay((Widget *)GetTopLevelWidget()),
                SWAP_LONG(*((WINDOW *)(Packet+4)),WhichTool->ConnectPrefix.byteOrder),
                SWAP_LONG(WhichTool->LocalRootInfo.windowId,WhichTool->ConnectPrefix.byteOrder),
                Win_Attr.x, Win_Attr.y, &abs_x, &abs_y, &children);
            Win_Attr.x = abs_x;
            Win_Attr.y = abs_y;
         }

#       if DEBUGHTOS_X11_PopUp_Packet
         fprintf( stderr, "HTOS_X11Packet(): PopUpInfo Absolute Window %d: Win_Attr.x  %d and Win_Attr.y %d \n",
             SWAP_LONG(*((WINDOW *)(Packet+4)),WhichTool->ConnectPrefix.byteOrder), Win_Attr.x, Win_Attr.y);
#       endif

         if ( WhichTool->TokenHolder == GetLocalSubID() ) {
#       if DEBUGHTOS_X11_PopUp_Packet
            fprintf( stderr, "HTOS_X11Packet():I have the token, modify x and y relative to me \n");
#       endif
            *((INT16 *)(Packet+8)) = SWAP_INT(SWAP_INT(*((INT16 *)(Packet+8)),WhichTool->ConnectPrefix.byteOrder) -
                Win_Attr.x,WhichTool->ConnectPrefix.byteOrder);
            *((INT16 *)(Packet+10)) = SWAP_INT(SWAP_INT(*((INT16 *)(Packet+10)),WhichTool->ConnectPrefix.byteOrder) -
                Win_Attr.y,WhichTool->ConnectPrefix.byteOrder);

            if (SendData(Port, Packet, sizeof(PopUpInfo)) == ERROR) {
#       if DEBUGHTOS_X11_PopUp_Packet
               fprintf( stderr, "HTOS_X11Packet():ERROR in sending PopUpReply Packet back to host");
#       endif
               CleanUpTool(WhichTool);
               return;
            }
         }
         FreeMem( ToolID );
         FreeMem( Length );
         FreeMem( Packet );
         return;
      }
      if ((WhichTool->popup)&&( *Packet == 1) && ExpectPopUp ) {
         ExpectPopUp = FALSE;

#       if DEBUGHTOS_X11_PopUp_Packet
         fprintf( stderr, "HTOS_X11Packet():A PopUp Window Request received:\n");
#       endif

#       if DEBUGHTOS_X11_PopUp_Packet
         fprintf( stderr,"LocalToolWindowPosition:Win_Attr.x %d, Win_Attr.y %d\n",
             Win_Attr.x,Win_Attr.y);
         fprintf( stderr, "HTOS_X11Packet(): relative x and y of popup are %d , %d \n", 
             *((INT16 *)(Packet+ 12)), *((INT16 *)(Packet+14)));

#   endif
         *((INT16 *)(Packet+12)) = SWAP_INT(SWAP_INT(*((INT16 *)(Packet+12)),WhichTool->ConnectPrefix.byteOrder) +
             Win_Attr.x,WhichTool->ConnectPrefix.byteOrder);
         *((INT16 *)(Packet+14)) = SWAP_INT(SWAP_INT(*((INT16 *)(Packet+14)),WhichTool->ConnectPrefix.byteOrder) +
             Win_Attr.y,WhichTool->ConnectPrefix.byteOrder);
#       if DEBUGHTOS_X11_PopUp_Packet
         fprintf( stderr, "HTOS_X11Packet(): Absolute x and y of popup are %d , %d \n", 
             *((INT16 *)(Packet+12)), *((INT16 *)(Packet+14)));
#       endif
      }
/*---------------------------------------------------------------
**
** Begin Kantsios addition:
*/
 
if (WhichTool->seqnum){
   ++(WhichTool->ServerRequestNumber);
   (void) AddToSNTable(WhichTool, WhichTool->ServerRequestNumber , *TheSeqNum);
#       if DEBUGHTOS_X11Packet
   fprintf(stderr, " Adding to Table: (%d, %d )\n", WhichTool->ServerRequestNumber , *TheSeqNum );
   fflush( stderr );
# endif
}
 
/*
** Before sending the whole mess off to the server, we need
** to increment server's sequence number, store the sequence
** number combination into the circular queue, and then send it
** to the server.
**
**  End Kantsios addition
**-------------------------------------------------------------*/

      if((returnVal=TimedSendData( WhichTool->OutConnection, Packet, *Length ))==ERROR) CleanUpTool(WhichTool);
      else if (returnVal == TIMEOUT) {
         if ( SendByte(GetHostConnection(), SH_StopTool)==ERROR){
            EndSession();
            return;
         }
         if ( SendByte(GetHostConnection(), WhichTool->ToolID)==ERROR){
            EndSession();
            return;
         }
         TimeOutFlag = TRUE;
         CleanUpTool(WhichTool);
         TimeOutFlag = FALSE;
      }

   }

   FreeMem( ToolID );
   FreeMem( Length );
   FreeMem( Packet );
   if ( WhichTool->CatchUpMode == 1 )
      if (SendByte(GetHostConnection(),SH_SendArchivePacket)==ERROR) EndSession();


}


/*
 *   HTOS_InternalAtom()
 *
 *   Cook a Tool.
 *
 *   Protocol:   HS_InternalAtom | Tool ID | AtomValue | Length | AtomName
 *
 *      HS_InternalAtom      Protocol ID Tag
 *      ToolID         Tool ID of sender
 *      AtomValue      32-bit host atom value in tool's byte order
 *      Length         16-bit name length in tool's byte order
 *      AtomName      Name of atom, "Length" bytes.
 *
 *   RETURN VALUE:   void
 */
void HTOS_InternalAtom( Port, Who )
Socket Port;
Subscriber *Who;
{
   BYTE      *ToolID;
   Tool      *WhichTool;
   CARD32      *AtomValue;
   CARD16      *Length;
   BYTE      *AtomName;
   CARD32      LocalAtomValue;
   int      i;

#   if DEBUGHTOS_InternalAtom
   fprintf( stderr, "HTOS_InternalAtom()\n" );
#   endif

   ToolID = (BYTE *)ReadData( Port, sizeof(BYTE), BLOCK );
   if ( (WhichTool=FindToolByID(*ToolID)) == NULL ) {
      /* xyzzy - for now trash this packet, it should't have
      arrived to us in the first place, the efficient fix
      is to let the host rcognize this and do't send it to us */
#   if DEBUGHTOS_InternalAtom
      fprintf(stderr,  "HTOS_InternalAtom(): Tool ID %d unknown.", *ToolID );
#   endif
      FreeMem(ToolID);
      TrashIt(Port);
      return;
   }
   AtomValue = (CARD32 *)ReadData( Port, sizeof(CARD32), BLOCK );
   Length = (CARD16 *)ReadData( Port, sizeof(CARD16), BLOCK );
   *Length = SWAP_INT(*Length,WhichTool->ConnectPrefix.byteOrder);
#   if DEBUGHTOS_InternalAtom
   fprintf( stderr, "HTOS_InternalAtom(): Receiving string with %d bytes\n", *Length );
#   endif
   AtomName = (BYTE *)ReadData( Port, *Length, BLOCK );
   AtomName[*Length] = NULL;
#   if DEBUGHTOS_InternalAtom
   fprintf( stderr, "HTOS_InternalAtom(): Host's atom %08x = '%s'\n", (XID)(*AtomValue),
       AtomName );
#   endif

   /* Grab the value of the atom from the local display server */

   LocalAtomValue = XInternAtom( XtDisplay((Widget *)GetTopLevelWidget()),
       (char *)AtomName, FALSE );

   LocalAtomValue = SWAP_LONG( LocalAtomValue, WhichTool->ConnectPrefix.byteOrder );

   /* Add it to the list, but put it at the HEAD of the list to make the
    * resource ID translator run at its best.
    */

   WhichTool->VisualIDList = (XID *)realloc( WhichTool->VisualIDList,
       (++(WhichTool->NumVisualIDs) * sizeof(XID) ) );
   for ( i=(WhichTool->NumVisualIDs-1); i > 0; i-- ) {
      WhichTool->VisualIDList[i] = WhichTool->VisualIDList[i-1];
   }
   WhichTool->VisualIDList[0] = *AtomValue;

   WhichTool->LocalVisualIDList = (XID *)realloc( WhichTool->LocalVisualIDList,
       (++(WhichTool->LocalNumVisualIDs) * sizeof(XID) ) );
   for ( i=(WhichTool->LocalNumVisualIDs-1); i > 0; i-- ) {
      WhichTool->LocalVisualIDList[i] = WhichTool->LocalVisualIDList[i-1];
   }
   WhichTool->LocalVisualIDList[0] = LocalAtomValue;

#   if DEBUGHTOS_InternalAtom
   fprintf( stderr, "HTOS_InternalAtom(): Host's atom %08x --> %08x\n", (XID)(*AtomValue),
       LocalAtomValue );
#   endif

   /* Clean up and go home */

   FreeMem( ToolID );
   FreeMem( AtomValue );
   FreeMem( Length );
   FreeMem( AtomName );
   if( WhichTool->CatchUpMode == 1 ) if (SendByte(GetHostConnection(),SH_SendArchivePacket)==ERROR) EndSession();

}


/*
 *      HTOS_AllocColor()
 *
 *      Protocol:   xtv_ACP packet format as defined in xspecial.h
 *
 *      1       HS_AllocColor           Protocol ID Tag
 *      1       ToolID                  Tool ID of sender
 *      2       Red         Recvd in tool byte order 
 *   2   Green         ditto
 *   2   Blue         ditto
 *   4   Cmap         Colormap  -- see caveat
 *      4       PixelValue              Recvd in Tool byteorder
 *
 *      RETURN VALUE:   void
 *
 *  Modified : FSH
 *
 */
void HTOS_AllocColor( Port, Who )
Socket Port;
Subscriber *Who;
{

   BYTE            *ToolID;
   BYTE      *Packet;
   Tool            *WhatTool;
   XColor          x_color;
   xtv_ACP      xtv_ACPacket;

   ToolID = (BYTE *)ReadData( Port, sizeof(BYTE), BLOCK );
   if ( (WhatTool=FindToolByID(*ToolID)) == NULL ) {
#if DEBUGHTOS_AllocColor
      fprintf(stderr,"HTOS_AllocColor: Tool ID %d unknown.", *ToolID );
#endif
      FreeMem(ToolID);
      TrashIt(Port);
      return;
   }
   /*
   ** Now read the rest in format we defined
   */
   Packet = (BYTE *) ReadData( Port, sizeof(xtv_ACP) - 2,BLOCK);
   memcpy(&xtv_ACPacket.red,Packet,sizeof(xtv_ACP) - 2);
   /*
   ** swap values and setup 
   */
   x_color.red = SWAP_INT(xtv_ACPacket.red,
       WhatTool->ConnectPrefix.byteOrder);
   x_color.green = SWAP_INT(xtv_ACPacket.green,
       WhatTool->ConnectPrefix.byteOrder);
   x_color.blue = SWAP_INT(xtv_ACPacket.blue,
       WhatTool->ConnectPrefix.byteOrder);
   /*
   ** Translate color map 
   */
   xtv_ACPacket.cmap = TranslateForward(WhatTool,xtv_ACPacket.cmap,
       S_XLATE);
   xtv_ACPacket.cmap = SWAP_LONG(xtv_ACPacket.cmap,
       WhatTool->ConnectPrefix.byteOrder);


   /*
   ** Now make the local call 
   */

#if DEBUGHTOS_AllocColor
   fprintf( stderr, "HTOS_AllocColor(): Calling XAllocColor\n"); 
   fprintf( stderr, "HTOS_AllocColor(): cmap 0x%08x red %d green %d blue %d\n",
                      xtv_ACPacket.cmap, x_color.red, x_color.green, x_color.blue); 
#endif

   XAllocColor(XtDisplay((Widget *)GetTopLevelWidget()),
       xtv_ACPacket.cmap, &x_color);
   /*
   ** both list items need to be in consistent order
   */
   x_color.pixel = SWAP_LONG(x_color.pixel,
       WhatTool->ConnectPrefix.byteOrder);

   /*
   ** Now add the pixel's to the list and wrap it up
   */
   add_pixels_to_lists(WhatTool,xtv_ACPacket.pixel,x_color.pixel);
   FreeMem( ToolID );
   FreeMem( Packet);
#if DEBUGHTOS_AllocColor
   fprintf( stderr, 
       "HTOS_AllocColor() localPix = %d\n", x_color.pixel );
#endif

   if ( WhatTool->CatchUpMode == 1 )
      if (SendByte(GetHostConnection(),SH_SendArchivePacket)==ERROR)
         EndSession();


}



/*
 *      HTOS_AllocNamedColor() - 
 *
 *      
 *
 *      Protocol:   
 *
 *      1       HS_AllocNamedColor   Protocol ID Tag
 *      1       toolID         Tool ID of sender
 *   2   name length      Color name length in bytes
 *   4   cmap         Colormap
 *      4       pixel         pixel from remote host
 *   n   Name         actual name to lookup
 *
 *      RETURN VALUE:   void
 *
 *  Modified : FSH
 *
 */

void HTOS_AllocNamedColor( Port, Who )
Socket Port;
Subscriber *Who;
{

   BYTE            *ToolID;
   Tool            *WhichTool;
   CARD16      *NameLen;
   Colormap        *CMap;
   CARD32      *RemotePixel;
   BYTE      *ColorName;
   XColor          x_color;
   XColor          RgbDbDef;

#if DEBUGHTOS_AllocNamedColor
   fprintf( stderr,  "HTOS_AllocNamedColor()\n" );
#endif

   ToolID = (BYTE *)ReadData( Port, sizeof(BYTE), BLOCK );
   if ( (WhichTool=FindToolByID(*ToolID)) == NULL ) {
#if DEBUGHTOS_AllocNamedColor
      fprintf(stderr,"HTOS_AllocNamedColor: Tool ID %d unknown.", *ToolID );
#endif
      FreeMem(ToolID);
      TrashIt(Port);
      return;
   }

   /*
   ** Get the name length
   */
   NameLen = (CARD16 *) ReadData( Port, sizeof(CARD16), BLOCK);
   *NameLen = SWAP_INT(*NameLen, WhichTool->ConnectPrefix.byteOrder);
   /*
   ** Now the CMAP
   */
   CMap = (Colormap *) ReadData( Port, sizeof(Colormap), BLOCK);
   *CMap = TranslateForward(WhichTool,*CMap, S_XLATE);
   *CMap = SWAP_LONG(*CMap,WhichTool->ConnectPrefix.byteOrder);

   /*
   ** Now the pixel 
   */
   RemotePixel = (CARD32 *) ReadData( Port, sizeof(CARD32), BLOCK);

   /*
   ** And the name 
   */
   ColorName =(BYTE *)ReadData( Port, *NameLen, BLOCK );
#if DEBUGHTOS_AllocNamedColor
   fprintf(stderr,"HTOS_AllocNamedColor: name |%s| :len = %d :rpixel = %ld\n", 
       ColorName,*NameLen & 0x00ffff,*RemotePixel);
#endif   
   /*
   ** Now the local call
   */
   XAllocNamedColor(XtDisplay((Widget *)GetTopLevelWidget()), 
       *CMap, (char *)ColorName,&x_color, &RgbDbDef);

   /*
   ** All list items need to be in consistent tool order
   */
   x_color.pixel = SWAP_LONG(x_color.pixel,
       WhichTool->ConnectPrefix.byteOrder);

   /*
   ** Add this pixel to translate list 
   */

   add_pixels_to_lists(WhichTool,*RemotePixel,x_color.pixel);
   /*
   ** Free up buffers
   */
   FreeMem(ToolID);
   FreeMem(NameLen);
   FreeMem(CMap);
   FreeMem(RemotePixel);
   FreeMem(ColorName);


   if ( WhichTool->CatchUpMode == 1 )
      if (SendByte(GetHostConnection(),SH_SendArchivePacket)==ERROR)
         EndSession();


}


/*
 *   HTOS_YouAre()
 *
 *   Cook a Tool.
 *
 *      Protocol:       HS_YouAre | SubscriberID | HostDisplayNumber
 *
 *              HS_YouAre               Protocol ID Tag
 *              SubscriberID            Local subscriberID
 *              HostDisplayNumber       To start tool remotely
 *
 *   RETURN VALUE:   void
 */
void HTOS_YouAre( Port, Who )
Socket Port;
Subscriber *Who;
{
   BYTE      WhoAmI;

   WhoAmI = GetByte( Port );

   SetLocalSubID( WhoAmI );
   HostDisplayNumber = GetByte(Port);
#       if DEBUGHTOS_YouAre
   fprintf( stderr, "HTOS_YouAre(): WhoAmI: %d , HostDisplayNumber %d \n", 
       WhoAmI, HostDisplayNumber);
#       endif


}


/*
 *   HTOS_TakeToken()
 *
 *   Cook a Tool.
 *
 *   Protocol:   HS_TakeToken | ToolID | SubscriberID
 *
 *      HS_TakeToken      Protocol ID Tag
 *      ToolID         Tool ID of tool in question
 *      SubscriberID      Local subscriberID
 *
 *   RETURN VALUE:   void
 */
void HTOS_TakeToken( Port, Who )
Socket Port;
Subscriber *Who;
{
   BYTE      WhoHasIt;
   Subscriber   *WhoIsIt;
   BYTE      ToolID;
   Tool      *WhatTool;

   ToolID = GetByte( Port );
   WhoHasIt = GetByte( Port );

   if ( (WhatTool=FindToolByID(ToolID)) == NULL ) {
      /* xyzzy - for now trash this packet, it should't have
      arrived to us in the first place, the efficient fix
      is to let the host rcognize this and do't send it to us */
#       if DEBUGHTOS_TakeToken
      fprintf(stderr,  "HTOS_TakeToken(): Tool ID %d unknown.", ToolID );
#   endif
      return;
   }


   /* Give Token of every tool in the group */
   SwitchTokenGroup(WhatTool->ToolGID, WhoHasIt);

   WhoIsIt = FindSubscriberByID( WhoHasIt );
   if (GetLocalSubID() == WhoHasIt) {
      WhatTool->AskedToGetToken = FALSE;
      UpdateMessage( "Now you have the token for %s\n", WhatTool->ToolCommand);
   } else {

#   if DEBUGHTOS_TakeToken
      fprintf( stderr, "HTOS_TakeToken(): %s now has the token for %s\n",
          (WhoHasIt == HOST_ID ? "THE HOST" : WhoIsIt->LoginName),
          WhatTool->ToolCommand );
#   endif

      UpdateMessage( "%s has the token for %s\n", ( WhoHasIt == HOST_ID ? "The Chairman" : WhoIsIt->LoginName ),
          WhatTool->ToolCommand );
   }
}


/*
 *   HTOS_ForceTokenRelease()
 *
 *   Cook a Tool.
 *
 *   Protocol:   HS_ForceTokenRelease | ToolID
 *
 *      HS_ForceTokenRelease   Protocol ID Tag
 *      ToolID         Tool ID of tool in question
 *
 *
 *   RETURN VALUE:   void
 */
void HTOS_ForceTokenRelease( Port, Who )
Socket Port;
Subscriber *Who;
{
   BYTE      WhoHasIt;
   Subscriber   *WhoIsIt;
   BYTE      ToolID;
   Tool      *WhatTool;

   ToolID = GetByte( Port );

   if ( (WhatTool=FindToolByID(ToolID)) == NULL ) {
#       if DEBUGHTOS_ForceTokenRelease
      fprintf(stderr, "HTOS_ForceTokenRelease:  Tool ID %d unknown.", ToolID );
#       endif
      return;
   }


#   if DEBUGHTOS_ForceTokenRelease
   fprintf( stderr, "HTOS_ForceTokenRelease(): Host snatched %s token.\n",
       WhatTool->ToolCommand );
#   endif

   UpdateMessage( "The chairman has snatched\n the token for tool %s", WhatTool->ToolCommand );

   /* Give Token of every tool in the group */
   SwitchTokenGroup(WhatTool->ToolGID, HOST_ID);
}

/*
 *   HTOS_AskTokenRelease()
 *
 *
 *   Protocol:   HS_AskTokenRelease | SubscriberID | ToolID
 *
 *      HS_AskTokenRelease   Protocol ID Tag
 *      SubscriberID      Subscriber ID that needs token
 *      ToolID         Tool ID of tool in question
 *
 *
 *   RETURN VALUE:   void
 */
void HTOS_AskTokenRelease( Port, Who )
Socket Port;
Subscriber *Who;
{
   Subscriber   *WhoIsIt;
   BYTE      ToolID;
   BYTE      SubscriberID;
   Tool      *WhatTool;

   SubscriberID = GetByte( Port );
   ToolID = GetByte( Port );

   if ( (WhoIsIt = FindSubscriberByID( SubscriberID )) == NULL){
#       if DEBUGHTOS_AskTokenRelease
      printf(stderr, "HTOS_AskTokenRelease(): Subscriber ID %d unknown.", SubscriberID );
#       endif
      return;
   }


   if ( (WhatTool=FindToolByID(ToolID)) == NULL ) {
#       if DEBUGHTOS_DeleteTool
      fprintf(stderr, "HTOS_AskTokenRelease:  Tool ID %d unknown.", ToolID );
#       endif
      return;
   }


#   if DEBUGHTOS_AskTokenRelease
   fprintf( stderr, "HTOS_AskTokenRelease(): Participant %s needs %s token.\n",
       WhoIsIt->LoginName, WhatTool->ToolCommand );
#   endif

   if ( WhatTool->TokenHolder == GetLocalSubID() ) {
      UpdateMessage( "%s wants the token\n for tool %s.\n Select %s tool and press [Drop Token]", WhoIsIt->LoginName,
          WhatTool->ToolCommand, WhatTool->ToolCommand);
   }

}
TrashIt(Port)
Socket Port;
{
   CARD32      *Length;
   BYTE      *Packet;

   Length = (CARD32 *)ReadData( Port, sizeof(CARD32), BLOCK );
   Packet = (BYTE *)ReadData( Port, *Length, BLOCK );
   FreeMem( Length );
   FreeMem( Packet );
}
/*
 *      HTOS_CreatorTakeToken()
 *
 *      Cook a Tool.
 *
 *      Protocol:       HS_CreatorTakeToken | ToolID | SubscriberID
 *
 *              HS_CreatorTakeToken            Protocol ID Tag
 *              ToolID                  Tool ID of tool in question
 *              SubscriberID            Local subscriberID
 *
 *      RETURN VALUE:   void
 */
void HTOS_CreatorTakeToken( Port, Who )
Socket Port;
Subscriber *Who;
{
   BYTE            WhoHasIt;
   Subscriber      *WhoIsIt;
   BYTE            ToolID;
   Tool            *WhatTool;

   ToolID = GetByte( Port );
   WhoHasIt = GetByte( Port );

   if ( (WhatTool=FindToolByID(ToolID)) == NULL ) {
      /* xyzzy - for now trash this packet, it should't have
                arrived to us in the first place, the efficient fix
                is to let the host rcognize this and do't send it to us */
#       if DEBUGHTOS_CreatorTakeToken
      fprintf(stderr,  "HTOS_CreatorTakeToken(): Tool ID %d unknown.", ToolID );
#       endif
      return;
   }

   /* if I am the owner and  did't ask for it, it means no one needs it and  I 
should keep it */

   UpdateMessage( "No one wants %s token \nYou have it.\nsince you are the creator of this tool" , WhatTool->ToolCommand);
   /* Give Token of every tool in the group */
   SwitchTokenGroup(WhatTool->ToolGID, WhoHasIt);

}



/*
 *      HTOS_CreatorTakeToken()
 *
 *      Cook a Tool.
 *
 *      Protocol:       HS_CreatorTakeToken | ToolID | SubscriberID
 *
 *              HS_CreatorTakeToken            Protocol ID Tag
 *              ToolID                  Tool ID of tool in question
 *              SubscriberID            Local subscriberID
 *
 *      RETURN VALUE:   void
 */
void HTOS_ContinueTakeToken( Port, Who )
Socket Port;
Subscriber *Who;
{
   BYTE            WhoHasIt;
   Subscriber      *WhoIsIt;
   BYTE            ToolID;
   Tool            *WhatTool;

   ToolID = GetByte( Port );
   WhoHasIt = GetByte( Port );

   if ( (WhatTool=FindToolByID(ToolID)) == NULL ) {
      /* xyzzy - for now trash this packet, it should't have
                arrived to us in the first place, the efficient fix
                is to let the host rcognize this and do't send it to us */
#       if DEBUGHTOS_ContinueTakeToken
      fprintf(stderr,  "HTOS_ContinueTakeToken(): Tool ID %d unknown.", ToolID );
#       endif
      return;
   }

   /* if I am the token holder and  no one needs it and  I should keep it */

   UpdateMessage("No one waiting for [%s] token.\nYou may keep it", WhatTool->ToolCommand);

   /* Give Token of every tool in the group */
   SwitchTokenGroup(WhatTool->ToolGID, WhoHasIt);
}



/*
 *   EndSession()
 *
 *   The chairman has terminated unexpectedly. Clean up NEATLY and
 *   exit.
 */

void EndSession()
{
   extern Widget   WSubscriber;
   extern Widget   WTool;
   Arg      args[5];
   int      n=0;

   XtSetArg( args[n],      XtNsensitive,           FALSE );
   n++;
   XtSetValues( WSubscriber, args, n );
   XtSetValues( WTool, args, n );

   CloseAllTools();

   UpdateMessage( "The Host has unexpectedly terminated.\n Press [Exit XTV] to exit." );
}


/*
 * Begin Murray Addition
 */

/*
 * HTOS_QueuedList()
 *
 * Get the list of queued ids from the socket
 *
 * 2-VII-91 - Changed to now accept a list of tool holders at the same time.
 *
 * Protocol:
 *     HS_QueuedList - Protocol ID Tag
 *     destination   - Specifies the routine to call with the token information.
 *     toolID        - Number of the tool that this is the info for. For checking consistency.
 *     numQueued     - Number of queued entries in the list
 *     <queuedEntries> - Each entry in the list of people waiting for the token.
 *
 * Local vars:
 *     numQueued     - (BYTE) holds the number of queued participants. Since SubscriberIDs are
 *                     are this size. There can't be anymore of these than numQueued can represent.
 *     queuedEntries - (BYTE[512]) holds the actual queued entries as they are received.
 *     queueIndex    - (short) used to index into the queuedEntry array.
 *     tIToolID      - (BYTE) holds the id of the tool. This is passed to tokenInfoSubscr() as
 *                     part of a consistency check.
 *
 * Return value: void
 */

XtCallbackProc target[] = {
   (XtCallbackProc) tokenInfoBuild, (XtCallbackProc) particInfoBuild};

void HTOS_QueuedList(Port, Who)
Socket Port;
Subscriber *Who;
{
   BYTE    destination;
   BYTE    numQueued;
   BYTE    queuedEntries[512];
   short   queueIndex;
   BYTE    tIToolID;

   /* 
   * Pull off the fixed part of the return message and set the appropriate local vars.
   */
   destination = GetByte(Port);
   tIToolID = GetByte(Port);
   numQueued = GetByte(Port);

   /*
   * Build the actual queuedEntries array by pulling ids off of the port given to us
   * as a parameter.
   */
   for (queueIndex=0; queueIndex < (short) numQueued; queueIndex++)
      queuedEntries[queueIndex] = GetByte(Port);

   (target[(int) destination])(tIToolID, (short) numQueued, queuedEntries);

   /*
   * tokenInfoBuild(tIToolID, (short) numQueued, queuedEntries);
   */
}

void HTOS_IncomingTools(Port, Who)
Socket Port;
Subscriber *Who;
{
   BYTE expectedNum, expectedLen;
   int  numTools, currentLen;
   char toolMsgBuffer[2048];
   char tmpName[128];

   expectedNum = GetByte(Port);
   if (!expectedNum)
   {
      (void) strcpy(toolMsgBuffer, "-- Expect zero tools from the host. --\n\r");
      UpdateMessage(toolMsgBuffer);
   }
   else
   {
      (void) strcpy(toolMsgBuffer, "Expect the following tools from the host:\n\r");
      for (numTools = 0; numTools < (int) expectedNum; numTools++)
      {
         (void) strcat(toolMsgBuffer, "  ");
         expectedLen = GetByte(Port);
         for (currentLen = 0; currentLen < (int) expectedLen; currentLen++)
            tmpName[currentLen] = (char) GetByte(Port);
         tmpName[currentLen] = '\0';
         (void) strcat(toolMsgBuffer, tmpName);
         (void) strcat(toolMsgBuffer, "\n\r");
      }
      /*
       * NOTE: unlockMessages is sent when the subscriber confirms/(denies) the
       * subscription to the last tool.
       *
       * 10-VII-91
       */
      UpdateMessage(toolMsgBuffer);
      lockMessages();
   }
}

void HTOS_EndIncoming(Port, Who)
Socket Port;
Subscriber *Who;
{
   /*
   * Now that we are at the end of the tools, we
   * want the message box to stay with new messages.
   */

   unlockMessages();
}

/*
 * End Murray Addition
 */

void HTOS_ToolsToJoin(Port, Who)
Socket Port;
Subscriber *Who;
{
   HandleToolsToJoinList();
}

void HTOS_CatchUpStart(Port, Who)
Socket Port;
Subscriber *Who;
{

   CatchUp = 1;
}

void HTOS_CatchUpEnd(Port, Who)
Socket Port;
Subscriber *Who;
{
   BYTE *id;
   Tool *tool;

   id = (BYTE *)ReadData( Port, sizeof(BYTE), BLOCK );
   if( (tool = FindToolByID(*id)) != NULL )
      tool->CatchUpMode = 0;
}
/* wang_x add */
void HTOS_ResizeWindowFunc ( Port, Who)
Socket Port;
Subscriber *Who;

{
   BYTE    *ToolID;
   Tool    *RezTool;
   ResizeInfo *ResData;
   Window   re_window;
   INT16     re_width;
   INT16     re_height;

   XWindowAttributes  WinAttr;


#  if DEBUGHTOS_ResizeWindow 
   fprintf(stderr, "HTOS_ResizeWindow: Staring .....\n");
#  endif
   ToolID =(BYTE *)ReadData( Port, sizeof(BYTE), BLOCK);
   RezTool = FindToolByID( *ToolID);
   ResData = (ResizeInfo *)ReadData( Port, sizeof(ResizeInfo), BLOCK);

   ResData->re_window = TranslateForward( RezTool, ResData->re_window, S_XLATE );

   re_width = SWAP_INT(ResData->re_width,RezTool->ConnectPrefix.byteOrder);
   re_height = SWAP_INT(ResData->re_height, RezTool->ConnectPrefix.byteOrder);
   re_window = SWAP_LONG(ResData->re_window, RezTool->ConnectPrefix.byteOrder);

#  if DEBUGHTOS_ResizeWindow
   fprintf(stderr, "HTOS_ResizeWindow: 0x%08x  %dx%d\n", 
                               re_window, re_width, re_height);
#  endif

   if (re_window != 0 && re_width !=0 && re_height != 0 && RezTool->resize){
      XResizeWindow(XtDisplay((Widget *)GetTopLevelWidget()),
                        re_window,re_width,re_height);
#  if DEBUGHTOS_ResizeWindow
      fprintf(stderr, "HTOS_ResizeWindow: Resized Window 0x%08x .....\n",
		re_window);
#  endif

   }
}
/* end wang_x add */



