/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

/*
**	xlspecial.c
**	Special routines for translating special packets
**
**	Modifications :  
**	FSH 4-SEP-91  Added mode to VListEntry 
**                      S_XLATE -> standard xlate
**                      PIXEL_XLATE -> Pixel xlate
**      FSH 4-SEP-91  Added mode parameter to all calls for RID translation.
**	FSH 24-SEP-91 Added XlateStoreNamedColors()
**	FSH 24-SEP-91 Added XlateFreeColors().
**	FSH 24-SEP-91 Added XlateGetWindowAttr() 
**      FSH 7-DEC-91  Changed remaining  calls to SwapInt() 
**                    and SwapLong() to Macros
**      FSH 11-DEC-91 Modified XlateFreeColors to remove pixels from 
**                    both color lists.
**      FSH 11-DEC-91 Turned Translate field on for backing pixel in 
**                    WinVTable
**	FSH 11-DEC-91 XlateGetWindowAttr() -  Backing pixel handled on reply. 
*/


#define	DEBUGSpecials	0&DEBUGON
#define DEBUGXlateFreeColors 0&DEBUGON 


#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#define NEED_REPLIES
#define NEED_EVENTS
#include <X11/Xproto.h>
#include "tool.h"
#include "xpacket.h"
#include "xlspecial.h"

#define	TRUE	1
#define	FALSE	0
extern  BYTE     XCOLOR;

typedef struct _VListEntry {
	char   FieldName[30];	/* Name of field for debugging */
	BOOL   Translate;       /* Whether or not it gets translated */
	BOOL   HandleZero;	/* Whether or not we handle the 0/1 case */
	int    mode;		/* FSH 4-SEP-91 mode to differ pixel from */
				/* visual id list and in future possibly  */
				/* to separate intern atom list 	  */
} VListEntry;

/* Definition of tables for translating LISTofVALUES entries in Window  Requests */

VListEntry   WinVTable[32] = {
	{ "BackgroundPixmap",	TRUE,	TRUE,	S_XLATE	},
	{ "BackgroundPixel",	TRUE,	FALSE,	PIXEL_XLATE },
	{ "BorderPixmap",	TRUE,	TRUE,	S_XLATE },
	{ "BorderPixel",	TRUE,	FALSE,	PIXEL_XLATE},
	{ "BitGravity",		FALSE,	FALSE,	S_XLATE},
	{ "WinGravity",		FALSE,	FALSE,	S_XLATE	},
	{ "BackingStore",	FALSE,	FALSE,	S_XLATE	},
	{ "BackingPlanes",	FALSE,	FALSE,	S_XLATE	},
	{ "BackingPixel",	TRUE,	FALSE,	PIXEL_XLATE},
	{ "OverrideRedirect",	FALSE,	FALSE,	S_XLATE	},
	{ "SaveUnder",		FALSE,	FALSE,	S_XLATE	},
	{ "EventMask",		FALSE,	FALSE,	S_XLATE	},
	{ "DNPMask",		FALSE,	FALSE,	S_XLATE	},
	{ "Colormap",		TRUE,	TRUE,	S_XLATE	},
	{ "Cursor",		TRUE,	TRUE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE },
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	}
};

/* Definition of tables for translating LISTofVALUES entries in GC Requests */

VListEntry   GCVTable[32] = {
	{ "Function",		FALSE,	FALSE,	S_XLATE	},
	{ "PlaneMask",		FALSE,	FALSE,	S_XLATE	},
	{ "Foreground",		TRUE,	FALSE,	PIXEL_XLATE }, 
	{ "Background",		TRUE,	FALSE,	PIXEL_XLATE }, 
	{ "LineWidth",		FALSE,	FALSE,	S_XLATE	},
	{ "LineStyle",		FALSE,	FALSE,	S_XLATE	},
	{ "CapStyle",		FALSE,	FALSE,	S_XLATE	},
	{ "JoinStyle",		FALSE,	FALSE,	S_XLATE	},
	{ "FillStyle",		FALSE,	FALSE,	S_XLATE	},
	{ "FillRule",		FALSE,	FALSE,	S_XLATE	},
	{ "Tile",		TRUE,	FALSE,	S_XLATE	},
	{ "Stipple",		TRUE,	FALSE,	S_XLATE	},
	{ "TileStippleXOrigin",	FALSE,	FALSE,	S_XLATE	},
	{ "TileStippleYOrigin",	FALSE,	FALSE,	S_XLATE	},
	{ "Font",		TRUE,	FALSE,	S_XLATE	},
	{ "SubWindowMode",	FALSE,	FALSE,	S_XLATE	},
	{ "GraphicsExposures",	FALSE,	FALSE,	S_XLATE	},
	{ "ClipXOrigin",	FALSE,	FALSE,	S_XLATE	},
	{ "ClipYOrigin",	FALSE,	FALSE,	S_XLATE	},
	{ "ClippingMask",	TRUE,	TRUE,	S_XLATE	},
	{ "DashOffset",		FALSE,	FALSE,	S_XLATE	},
	{ "Dashes",		FALSE,	FALSE,	S_XLATE	},
	{ "ArcMode",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,	FALSE,	S_XLATE	},
	{ "UNDEFINED",		FALSE,  FALSE,	S_XLATE	}
};




/*
 *   void FixValue
 *
 *   Translate the contents of a LISTofVALUE entry in a CreateGC
 *   or ChangeGC request based on the value of BitMask.
 *
 *   RETURN VALUE:   (void)
 */
void FixValue( BitMask, ValueList, Table, WhatTool )
CARD32 BitMask;
XID *ValueList;
VListEntry *Table;
Tool *WhatTool;
{
	int	BitNumber;
	int	WhereAt = 0;

/* chungg_bugfix */

	BitMask = SWAP_LONG(BitMask,WhatTool->ConnectPrefix.byteOrder);

	for ( BitNumber=0; BitNumber < 32; BitNumber++ ) {
		if ( BitMask & (1 << BitNumber) ) {
#			if DEBUGSpecials
			fprintf( stderr, "FixValue(): Field %s present",
			    Table[BitNumber].FieldName );
#			endif

			if ( (Table[BitNumber].Translate) && 
			    (!(Table[BitNumber].HandleZero) ||
/* chungg_bugfix */
			     !(SWAP_LONG(ValueList[WhereAt],WhatTool->ConnectPrefix.byteOrder) <= 1)) ) {
   
                                /* 
                                ** FSH 4-SEP-91 Add mode to avoid ID clash  
                                */
                                 ValueList[WhereAt] =
                                         TranslateForward( WhatTool,
                                                   ValueList[WhereAt],
                                                   Table[BitNumber].mode);

#				if DEBUGSpecials
				fprintf( stderr, "  -- TRANSLATED\n" );
#				endif
			} else {
#				if DEBUGSpecials
				fprintf( stderr, "\n" );
#				endif
			}
			WhereAt++;
		}
	}
}


/*
 *   void XlateCreateGC
 *
 *   Complex translation for CreateGC (55)
 *
 *   RETURN VALUE:   (void)
 */
void XlateCreateGC( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{
#	if DEBUGSpecials
	fprintf( stderr, "XlateCreateGC()\n" );
#	endif
	FixValue(	((xCreateGCReq *)Packet)->mask,		/* Bitmask */
	(XID *)(Packet+sizeof(xCreateGCReq)),	/* Pointer to ValueList */
	GCVTable,				/* Pointer to xlate table */
	WhatTool );				/* Pointer to tool structure */
}



/*
 *   void XlateChangeGC
 *
 *   Complex translation for ChangeGC (56)
 *
 *   RETURN VALUE:   (void)
 */
void XlateChangeGC( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{
#	if DEBUGSpecials
	fprintf( stderr, " XlateChangeGC()\n" );
#	endif
	FixValue(	((xChangeGCReq *)Packet)->mask,		/* Bitmask */
	(XID *)(Packet+sizeof(xChangeGCReq)),	/* Pointer to ValueList */
	GCVTable,				/* Pointer to xlate table */
	WhatTool );				/* Pointer to tool structure */
}



/*
 *   void XlateCreateWindow
 *
 *   Complex translation for CreateWindow (01)
 *
 *   RETURN VALUE:   (void)
 */
void XlateCreateWindow( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{

#	if DEBUGSpecials
	fprintf( stderr, "XlateCreateWindow()\n" );
#	endif
	FixValue(	((xCreateWindowReq *)Packet)->mask,		/* Bitmask */
	(XID *)(Packet+sizeof(xCreateWindowReq)),	/* Pointer to ValueList */
	WinVTable,					/* Pointer to xlate table */
	WhatTool );					/* Pointer to tool structure */

	/* Check if the depth of window to be created is supported by the server */

	if ( ((xCreateWindowReq *)Packet)->depth > WhatTool->LocalRootInfo.nDepths ) {
		((xCreateWindowReq *)Packet)->depth = WhatTool->LocalRootInfo.rootDepth;
	}
}

void XlateCreatePixmap( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{
#	if DEBUGSpecials
	fprintf( stderr, "XlateCreatePixmap()\n" );
#	endif
	/* Check if the depth of the pixmap to be created is supported by the server */

	if ( ((xCreateWindowReq *)Packet)->depth > WhatTool->LocalRootInfo.nDepths ) {
		((xCreateWindowReq *)Packet)->depth = WhatTool->LocalRootInfo.rootDepth;
	}

}


/*
 *   void XlateChangeWindowAttr
 *
 *   Complex translation for ChangeWindowAttributes (02)
 *
 *   RETURN VALUE:   (void)
 */
void XlateChangeWindowAttr( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{
#	if DEBUGSpecials
	fprintf( stderr, "XlateChangeWindowAttr()\n" );
#	endif
	FixValue(	((xChangeWindowAttributesReq *)Packet)->valueMask,	/* Bitmask */
	(XID *)(Packet+sizeof(xChangeWindowAttributesReq)),	/* Ptr to ValueList */
	WinVTable,						/* Ptr to xlate table */
	WhatTool );						/* Ptr to tool structure */

}


/*
 *   void XlateConfigWindow
 *
 *   Complex translation for ConfigureWindow (12)
 *
 *   RETURN VALUE:   (void)
 */
void XlateConfigWindow( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{
	xConfigureWindowReq	*Request = (xConfigureWindowReq *)Packet;
	XID			*AttrTable = (XID *)(Packet+sizeof(xConfigureWindowReq));
	int			i;
	CARD16			tmp_mask;

#	if DEBUGSpecials
	fprintf( stderr, "XlateConfigWindow(()\n" );
#	endif

/* chungg_bugfix */
	tmp_mask = SWAP_INT(Request->mask,WhatTool->ConnectPrefix.byteOrder);
	for ( i=0; i < 16; i++ ) {
		if ( (tmp_mask & (1 << i)) ) {
			AttrTable[i] = TranslateForward(WhatTool, 
							AttrTable[i],
 							S_XLATE );
                }
 	}
}


/*
 *   void XlateRotateProps
 *
 *   Complex translation for RotateProperties (114)
 *
 *   RETURN VALUE:   (void)
 */
void XlateRotateProps( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{
	xRotatePropertiesReq	*Request = (xRotatePropertiesReq *)Packet;
	XID			*AtomList = (XID *)(Packet + sizeof(xRotatePropertiesReq));
	int			i;
	CARD16			nAtoms;


#	if DEBUGSpecials
	fprintf( stderr, "XlateRotateProps()\n" );
#	endif

/* chungg_bugfix */
	nAtoms = SWAP_INT(Request->nAtoms,WhatTool->ConnectPrefix.byteOrder);
	for ( i=0; i < nAtoms ; i++ ) {
		AtomList[i] = TranslateForward( WhatTool, AtomList[i],S_XLATE);
	}
}



#define ROUNDUP(nbytes, pad)  ((((nbytes) + ((pad) - 1)) / (pad)) * (pad))

/*
 *   void XlatePutImage
 *
 *   Complex translation for PutImage (72)
 *
 *   RETURN VALUE:      (void)
 *
 */
void XlatePutImage( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{
    if (WhatTool->xcolor) {
        xPutImageReq    *Request = (xPutImageReq *)Packet;
        BYTE            *Data = (BYTE *)(Packet + sizeof(xPutImageReq));
        XImage          *image = (XImage *)malloc( sizeof(XImage) );
        CARD32          pixelValue;
        register int    j;
        register int    x, y;


        image->width = SWAP_INT(Request->width, 
				WhatTool->ConnectPrefix.byteOrder);
        image->height = SWAP_INT(Request->height, 
				WhatTool->ConnectPrefix.byteOrder);
        image->xoffset = SWAP_INT(Request->leftPad,
				WhatTool->ConnectPrefix.byteOrder);
        image->format = Request->format; 
        image->depth = Request->depth;
        image->byte_order = WhatTool->HostConnectInfo.imageByteOrder;
        image->bitmap_unit =  WhatTool->HostConnectInfo.bitmapScanlineUnit;
        image->bitmap_bit_order = WhatTool->HostConnectInfo.bitmapBitOrder;
        image->red_mask = image->green_mask = image->blue_mask = 0;

        if (image->format == ZPixmap) {
                register xPixmapFormat *fmt = WhatTool->LocalPixmapFormats;
                register int i;

                for (i = WhatTool->HostConnectInfo.numFormats + 1; --i; ++fmt)
                        if (fmt->depth == image->depth) {
                                image->bits_per_pixel = fmt->bitsPerPixel;
                                break;  /* actually break out of for-loop */
                        }
        }
        image->bitmap_pad = WhatTool->HostConnectInfo.bitmapScanlinePad;
        image->data = (char *)Data;
        if (image->format == ZPixmap)
                image->bytes_per_line =
                        ROUNDUP((image->bits_per_pixel * image->width),
                                                image->bitmap_pad) >> 3;
        else
                image->bytes_per_line =
                         ROUNDUP((image->width + image->xoffset),
                                                image->bitmap_pad) >> 3;
        image->obdata = NULL;
        _XInitImageFuncPtrs (image);

        for (x=0; x < image->width; x++)
                for (y =0; y < image->height; y++)
        {
			pixelValue = XGetPixel(image, x, y);
                        for (j = 0; j < WhatTool->NumColorCells; j++)
                                if (pixelValue == WhatTool->ColorCellList[j]) {
                                        pixelValue = WhatTool->LocalColorCellList[j];
/*wahab fix */
pixelValue = SWAP_LONG(pixelValue,
                                WhatTool->ConnectPrefix.byteOrder);

                                        break;
                                }
                        XPutPixel(image, x, y, pixelValue);
        }
    }
}



/*
 *   void XlateStoreColors
 *
 *   Complex translation for StoreColors (89)
 *
 *   RETURN VALUE:      (void)
 *
 */
void XlateStoreColors( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{
        xStoreColorsReq         *Request = (xStoreColorsReq *)Packet;
        CARD32                  *ColorItemList = (CARD32 *)(Packet + sizeof(xStoreColorsReq));
        int                     i;
        CARD16                  nColorItems;

#       if DEBUGSpecials
        fprintf( stderr, "XlateStoreColors()\n" );
#       endif

        if (Request->cmap == WhatTool->HostRootInfo.defaultColormap) {
           nColorItems = (SWAP_INT(Request->length, WhatTool->ConnectPrefix.byteOrder)-2) / 3;
           for ( i=0; i < nColorItems ; i++ ) {
                   ColorItemList[3*i] = TranslateForward( WhatTool,
		                             ColorItemList[3*i],PIXEL_XLATE );
           }
	}
}


/*
** XlateStoreNamedColor (90)
**
** 
** Translate Pixel Id  
**
** FSH 21-SEP-91
*/

  void
XlateStoreNamedColor( Packet, WhatTool )
 BYTE *Packet;
 Tool *WhatTool;
{

  xStoreNamedColorReq  *Request = (xStoreNamedColorReq *) Packet;

#if DEBUGSpecials
  fprintf( stderr, "XlateStoreNamedColors()\n" );
#endif

  if (Request->cmap == WhatTool->HostRootInfo.defaultColormap) 
    Request->pixel = TranslateForward( WhatTool,Request->pixel, PIXEL_XLATE);
 
}


/*
 *   void XlateQueryColors
 *
 *   Complex translation for QueryColors (91)
 *
 *   RETURN VALUE:      (void)
 *
 */
void XlateQueryColors( Packet, WhatTool )
BYTE *Packet;
Tool *WhatTool;
{
        xQueryColorsReq         *Request = (xQueryColorsReq *)Packet;
        CARD32                  *PixelList = (CARD32 *)(Packet + sizeof(xQueryColorsReq));
        int                     i;
        CARD16                  nPixels;

#       if DEBUGSpecials
        fprintf( stderr, "XlateQueryColors()\n" );
#       endif

        if (Request->cmap == WhatTool->HostRootInfo.defaultColormap) {
          nPixels = SWAP_INT(Request->length, WhatTool->ConnectPrefix.byteOrder) - 2;
           for ( i=0; i < nPixels ; i++ ) {
                   PixelList[i] = TranslateForward( WhatTool, PixelList[i],
		                           PIXEL_XLATE);
           }
         }
}



/*
** XlateFreeColors()
**
** CMAP has been Translated so xlate pixel and remove them from both lists
**
**  FSH 21-SEP-91
**
**  Modifications :  FSH 11-DEC-91   Remove pixels from Color lists too
**
*/
  void
XlateFreeColors( Packet, WhatTool )
 BYTE *Packet;
 Tool *WhatTool;
{

  xFreeColorsReq  *Request = (xFreeColorsReq *) Packet;
  CARD32 *pixel = (CARD32 *)(Packet + sizeof(xFreeColorsReq));
  CARD32 remote_pixel;
  CARD32 local_pixel;
  CARD16 length = 
            SWAP_LONG(Request->length,WhatTool->ConnectPrefix.byteOrder) - 3;
  register int i;


#if DEBUGXlateFreeColors
  fprintf(stderr, "XlateFreeColors()\n" );
#endif

  if (Request->cmap == WhatTool->HostRootInfo.defaultColormap) {
    for ( i = 0; i < length ; i++ ) {
/* wahab fix
      remote_pixel = SWAP_LONG(*(pixel + i),WhatTool->ConnectPrefix.byteOrder);
*/
      remote_pixel = *(pixel + i);
      *(pixel + i) =
          local_pixel = TranslateForward( WhatTool, *(pixel + i), PIXEL_XLATE);
      remove_pixels_from_lists(WhatTool,remote_pixel,local_pixel);
    }	 
  }   
}


void ForceDropDead()
{
#	if DEBUGSpecials
	fprintf( stderr, "ForceDropDead()\n" );
#	endif
}



/*----------------------------------------------------------------------*
 *									*
 * Following are the routines that perform complex replacement of XIDs  *
 * on reply from X server.						*
 * 									*
 *----------------------------------------------------------------------*/

void XlateQueryTree (buf,  WhatTool)
BYTE *buf;
Tool *WhatTool;

{
	short n; /* number of windows in children */
	XID *XID_ptr;
	int i;

/* chungg_bugfix */
	n = SWAP_INT( (short) *((short *)(buf + 16)), WhatTool->ConnectPrefix.byteOrder);
	XID_ptr = (XID *) (buf + 32);

	for (i = 0; i < n; i++) {
		*XID_ptr = TranslateBackward (WhatTool, *XID_ptr,S_XLATE);
		XID_ptr++;
	}
}

void XlateListProp (buf,WhatTool)
BYTE *buf;
Tool *WhatTool;

{
	short n; /* number of windows in children */
	XID *XID_ptr;
	int i;

/* chungg_bugfix */
	n = SWAP_INT( (short) *((short *)(buf + 8)), WhatTool->ConnectPrefix.byteOrder);
	XID_ptr = (XID *) (buf + 32);

	for (i = 0; i < n; i++) {
		*XID_ptr = TranslateBackward (WhatTool, *XID_ptr,S_XLATE);
		XID_ptr++;
	}
}

void XlateListFonts (buf, WhatTool)
BYTE *buf;
Tool *WhatTool;

{
	short m; /* number of FONTPROPs in properties */
	XID *XID_ptr;
	int i;

/* chungg_bugfix */
	m = SWAP_INT( (short) *((short *)(buf + 46)), WhatTool->ConnectPrefix.byteOrder);
	XID_ptr = (XID *) (buf + 64);

	for (i = 0; i < m; i++) {
		*XID_ptr = TranslateBackward (WhatTool, *XID_ptr,S_XLATE);
		XID_ptr++;
		XID_ptr++;
	}
}

void XlateColormaps (buf,WhatTool)
BYTE *buf;
Tool *WhatTool;

{
	short n; /* number of COLORMAPs in cmaps */
	XID *XID_ptr;
	int i;

	n = SWAP_INT( (short) *((short *)(buf + 8)), WhatTool->ConnectPrefix.byteOrder);
	XID_ptr = (XID *) (buf + 32);

	for (i = 0; i < n; i++) {
		*XID_ptr = TranslateBackward (WhatTool, *XID_ptr, S_XLATE);
		XID_ptr++;
	}
}



/*
** xlateGetWindowAttr
** 
** Reverse translate backing pixel ID 
** 
**  FSH 24-SEP-91
** 
**  Modifications:     FSH 11-DEC-91  
*/
  void 
XlateGetWindowAttr (Packet,WhatTool)
  BYTE *Packet;
  Tool *WhatTool;
{ 
  xGetWindowAttributesReply  *Reply = (xGetWindowAttributesReply *) Packet;
	
  Reply->backingPixel = TranslateBackward( WhatTool, 
                               Reply->backingPixel,PIXEL_XLATE); 
}
